module Main where

import Graphics.UI.GLUT -- hiding (position)
import Graphics.Rendering.OpenGL.GLU

import System
import System.Random
import Data.IORef
import Control.Monad
import qualified Data.Set as Set
import qualified Data.Map as Map

import Data.Fixed

-- TPP

import Model
import ViewSupport
import System.Exit
import Sprites

---------------
-- Global(s) --
---------------

timerInterval = 16 --16

----------------------
-- Type Definitions --
----------------------

type Point= Vertex3 GLdouble

data ViewPos = ViewPos Double Double Double

data SimState = Sim {
                  env::Env,
                  robot::Robot,
                  ctlrState::CtlrState,
                  pPos::ViewPos,
                  running::Bool,
                  count::Int,
                  keys::IORef (Set.Set Key)}

data SimRenderer=Renderer{rendererFunc::DisplayCallback,
                          sim::IORef SimState}

----------------------
-- Callback Support --
----------------------

addVer3::Num a => Vertex3 a -> Vertex3 a -> Vertex3 a
addVer3 (Vertex3 x1 y1 z1) (Vertex3 x2 y2 z2)
            = Vertex3 (x1+x2) (y1+y2) (z1+z2)

subVer3::Num a => Vertex3 a -> Vertex3 a -> Vertex3 a
subVer3 (Vertex3 x1 y1 z1) (Vertex3 x2 y2 z2)
            = Vertex3 (x1-x2) (y1-y2) (z1-z2)

addViewPos::ViewPos -> ViewPos  -> ViewPos
addViewPos (ViewPos a1 b1 r1) (ViewPos a2 b2 r2)
            = ViewPos ((a1+a2)`mod'`360) ((b1+b2)`mod'`360) (r1+r2)

subVec3::Point3D -> Point3D -> Point3D
subVec3 (x1,y1,z1) (x2,y2,z2)
            = ( (x1-x2),(y1-y2),(z1-z2) )

updateSim :: IORef SimState -> IO()
updateSim sim = do

    -- get simState info
    gs      <- readIORef sim 
    ks      <- readIORef $ keys gs
    oldPPos <- return $ pPos gs
    env     <- return $ env gs
    robotSt <- return $ robot gs
    cState  <- return $ ctlrState gs
    runF    <- return $ running gs
    count   <- return $ count gs

    -- get robot info
    (oldTheta,xx) <- return $ rDir  robotSt
    oldPos0  <- return $ rPos0 robotSt
    oldPos1  <- return $ rPos1 robotSt
    oldVel   <- return $ rVel  robotSt

    if (count `mod` 100) == 0 then print (count) else return ()
    if (count==800000) then exitWith ExitSuccess else return 0

    -- update controller,  only if not paused

    (newVel,dTheta,newCState) <- return $ 
          if runF then do
                      -- package up input and give to ctlr
                      (controlRobot CI {ciFront=(rFront robotSt),
                                        ciSide=(rSide robotSt),
                                        ciClk=count}
                                    cState)
                   else (oldVel,0.0,cState)

    -- update the robot itself

    newRobot <- return $ 
         if runF then do 
                     Robot {rPos0=(updatePosition oldPos0 (oldTheta,xx) newVel),
                            rPos1=oldPos0,
                            rDir=(oldTheta+dTheta,0.0), -- fake 2D
                            rVel=newVel,
                            rSide=(updateSideSensor oldPos0 (oldTheta,xx)),
                            rFront=(updateFrontSensor oldPos0 (oldTheta,xx)), -- hack .. make angle 2D
                            rSetP=0.0}
                  else robotSt

    -- for now nada
    newEnv <- return env

    -- handle keys
    viewPDeltas   <- return $ map (`kSearchMap` kVectorMap) $ Set.elems ks 
    sumOfDeltas   <- return $ foldl (addViewPos) (ViewPos 0 0 0) viewPDeltas
    ViewPos a b r <- return sumOfDeltas
    newPPos       <- return $ addViewPos oldPPos (ViewPos a b r) 
    
    newRunF <- return $ not (Set.member (Char ' ') ks)

    -- handle count .. only if not paused
    newCount      <- if runF then return $ (count+1)
                             else return $ count
    -- now update the Sim
    writeIORef sim 
      Sim{
            env       = newEnv,
            robot     = newRobot,
            ctlrState = newCState,
            pPos      = newPPos,
            running   = newRunF,
            count     = newCount,
            keys      = keys gs}

setPointOfView gs = do
  ViewPos alpha beta r <- return $ pPos gs
  let
   (x,y,z)    = calculatePointOfView alpha beta r
   (x2,y2,z2) = calculatePointOfView ((alpha+90)`mod'` 360) beta r
  lookAt (Vertex3 x y z) (Vertex3 0 0 0) (Vector3 x2 y2 z2)

calculatePointOfView alp bet r =
  let alpha =  alp*2*pi/fromIntegral 360
      beta = bet*2*pi/fromIntegral 360
      y = r * cos alpha
      u = r * sin alpha
      x = u * cos beta
      z = u * sin beta
  in (x,y,z)

-- Keyboard Related --
kSearchMap f ((ch,v):xs)= if f==ch then v else kSearchMap f xs
kSearchMap f []=ViewPos 0 0 0
kVectorMap = [

            (Char 'i',            ViewPos   0   0 (-0.01)),
            (Char 'o',            ViewPos   0   0   0.01),
            (SpecialKey KeyLeft,  ViewPos   0 359.5      0),
            (SpecialKey KeyRight, ViewPos   0   0.5      0),
            (SpecialKey KeyUp,    ViewPos   0.5   0      0),
            (SpecialKey KeyDown,  ViewPos 359.5   0      0)]
   
genTestCases = 
  map (\angle -> (angle, updateSideSensor (0.0,0.0,0.0) (angle,0.0))) 
   [1.0,11.0..351.0]

-----------------------
-- GL/GLUT Callbacks --
-----------------------

display simState tex1 tex2 tex3 = do

    loadIdentity
    clearColor $= Color4 0.0 0.0 0.0 0.0
    clear [ColorBuffer,DepthBuffer]    

    shadeModel $= Smooth
    lineSmooth $= Enabled

    blendEquation $= FuncAdd
    blendFunc $= (SrcAlpha, OneMinusSrcAlpha)
    textureFunction $= Replace


    gs <- readIORef simState

    baseDisp <- return $ Vector3 0.0 0.0 (0.0::Double) 
    
    setPointOfView gs    
    pointSize $=1.0
    currentColor $= Color4 0.89 0.474 0.137 1

    ------------------------------------------
    -- Draw Sensor Beams
    blend $= Enabled
    fBeamContext
    drawBeam (rPos0 (robot gs)) (rDir (robot gs)) (rFront (robot gs))

    sBeamContext
    (theta,_) <- return $ (rDir (robot gs))
    drawBeam (rPos0 (robot gs)) (theta-90.0,0) (rSide (robot gs))
    blend $= Disabled
    ------------------------------------------

    ---------------------------------------------
    -- Draw Robot
    robotContext
    showRobot (rPos0 (robot gs)) (rDir (robot gs))
    ---------------------------------------------

    ------------------------------------------
    -- Draw Driving Plane
    shellContext
    texture Texture2D $= Enabled    
    preservingMatrix $ do
       -- put plane top at z=0
       translate (Vector3 0.0 0.0 ((-0.025)::Double))
       cubeoidTex ([tex3]++(replicate 5 Nothing)) 6.0 6.0 0.05
    texture Texture2D $= Disabled    
    ------------------------------------------

    ------------------------------------------
    -- Draw Walls
    greyContext
    texture Texture2D $= Enabled 

    -- Robot field - Center of robot c,
    --  -2.3 <= c <= 2.3 in both x and y
    -- Outside these bounds => collision

    wallHgt  <- return (0.46::Double)
    wallThk  <- return (0.20::Double)    
    wallLen  <- return (5.60::Double)
    wallDisp <- return (2.50::Double)

    preservingMatrix $ do -- left wall
       translate (Vector3 (-(wallDisp+(wallThk/2.0))) 0.0 (wallHgt/2.0))
       rotateZ 90
       cubeoidTex (replicate 6 tex2) wallLen wallThk wallHgt

    preservingMatrix $ do -- right wall
       translate (Vector3   (wallDisp+(wallThk/2.0)) 0.0 (wallHgt/2.0))
       rotateZ 90
       cubeoidTex (replicate 6 tex2) wallLen wallThk wallHgt

    preservingMatrix $ do -- top wall
       translate (Vector3 0.0 (-(wallDisp+(wallThk/2.0))) (wallHgt/2.0))
       cubeoidTex  (replicate 6 tex2) wallLen wallThk wallHgt

    preservingMatrix $ do -- bottom wall
       translate (Vector3 0.0   (wallDisp+(wallThk/2.0)) (wallHgt/2.0))
       cubeoidTex  (replicate 6 tex2) wallLen wallThk wallHgt

    texture Texture2D $= Disabled    
    ------------------------------------------


    ------------------------------------------
    -- Draw Outer Sphere
    shellContext
    renderObject Wireframe (Sphere' 20 80 80)
    ------------------------------------------

    swapBuffers
    flush

    where 

        drawBeam (x,y,z) (xyDir,_) len = preservingMatrix $ do

           lenMax <- return 7.2
           translate $ Vector3 x y (0.2::Double)
           rotateZ (xyDir-0.0)
           rotateZ 180.0
           translate $ Vector3 (-len) 0.0 0.0
           rotateY 90.0  
           renderObject Solid (Cone (0.15 * (len / lenMax)) len 10 10) -- 4 9)


        showRobot (x,y,z) (xyDir,_) = preservingMatrix $ do

           texture Texture2D $= Enabled  

           rHt <-    return (0.3::Double)  
           rWidth <- return (0.4::Double)
           rLen <-   return (0.4::Double)

           currentColor $= Color4 0 1 0 1
           greyContext
           preservingMatrix $ do
              translate (Vector3 0 0 (rHt/2.0)) -- place on surface
              translate (Vector3 x y z)
              rotateZ (xyDir-90.0) -- hack since face rotated
              cubeoidTex ([tex1]++(replicate 6 tex2))  0.4 0.4 0.3

           texture Texture2D $= Disabled    

timerProc::IORef SimRenderer->IO ()
timerProc grRef = do
    gr <- readIORef grRef
    updateSim $ sim gr
    gs <- readIORef $ sim gr
    rendererFunc gr -- actually run the renderFunc .. i.e. display
    addTimerCallback timerInterval $ timerProc grRef


reshape simState screenSize@(Size w h) = do
  viewport $= ((Position 0 0), screenSize)
  matrixMode $= Projection
  loadIdentity
  let near   = 0.001
      far    = 40
      fov    = 30
      ang    = (fov*pi)/(360)
      top    = near / ( cos(ang) / sin(ang) )
      aspect = fromIntegral(w)/fromIntegral(h)
      right = top*aspect
  frustum (-right) right (-top) top near far
  matrixMode $= Modelview 0


keyboardProc keySet ch state _ _
    | (ch == Char ' ') && (state == Down) = toggleSpaceBar keySet ch
    | (ch == Char ' ') && (state == Up)   = return ()
    | ch     == Char 'q' = exitWith ExitSuccess 
    | state  == Down     = modifyIORef keySet (Set.insert ch)
    | state  == Up       = modifyIORef keySet (Set.delete ch)
    | otherwise          = return ()
       where
          toggleSpaceBar keyset ch = do
               ks <- readIORef keySet
               if (Set.member (Char ' ') ks) then modifyIORef keyset (Set.delete ch)
                                             else modifyIORef keyset (Set.insert ch)
----------
-- Main --
----------

main = do
    (progname, _) <- getArgsAndInitialize

    -- Sim
    keyState <- newIORef Set.empty

    simState <- newIORef Sim {
        env       = initEnv,
        robot     = initRobot,
        ctlrState = initCtlrState,
        pPos      = ViewPos 250 290 14.1, -- fov=30
        running   = True,
        count     = 0,
        keys      = keyState}


    -- GLUT
    initialDisplayMode $= [RGBAMode, DoubleBuffered, WithDepthBuffer,  WithAlphaComponent]

    initialWindowSize $= Size 800 800 -- 640 480
    
    createWindow "Robot Simulation"

    depthFunc $= Just Less
    autoNormal $= Disabled
    -- lighting features
    lighting $= Enabled
    light (Light 0) $= Enabled
    light (Light 1) $= Disabled
    position (Light 0) $= Vertex4 10 10 (-1) 10

    bTex <- preservingMatrix $ do
               createTexture "plate5.rgb" (False, False) 
    rTex <- preservingMatrix $ do
               createTexture "robot2.rgb" (False, False) 
    pTex <- preservingMatrix $ do
               createTexture "robotplain.rgb" (False, False) 
    dTex <- preservingMatrix $ do
               createTexture "dave2.rgb" (False, False) -- dave2 good

    simRenderer <- newIORef Renderer{

        rendererFunc=display simState rTex pTex bTex, sim=simState}

    displayCallback $= display simState rTex pTex bTex 
    reshapeCallback $= Just (reshape simState)
    keyboardMouseCallback $= Just (keyboardProc keyState)
    addTimerCallback timerInterval $ timerProc simRenderer
    
    mainLoop
    
