
module Model 

 where

import Data.Fixed

-----------------------
-- Types Definitions --
-----------------------

type Point3D = (Double, Double, Double)

type Point3DPair = (Point3D,Point3D)

data Robot = Robot {rPos0::Point3D,
                    rPos1::Point3D,
                    rDir::(Double,Double), -- theta,alpha
                    rVel::Double,
                    rSide::Double,
                    rFront::Double,
                    rSetP::Double}

data Env = Env {eWall::(Double,Double),
                eCycle::Int}

data CtlrInput = CI {ciFront::Double,
                     ciSide::Double,
                     ciClk::Int}

data CtlrState = CS {x1337::Bool}

----------------------
-- Support Routines --
----------------------

initRobot :: Robot
initRobot = Robot {rPos0=(0.0,0.0,0.0),
                   rPos1=(0.0,0.0,0.0),
                   rDir=(90.0,0.0), -- alpha (wrt +x), beta (wrt +z) [deg]
                   rVel=0.0,
                   rSide=0.0,
                   rFront=0.0,
                   rSetP=0.0}
initEnv :: Env
initEnv = Env     {eWall=(0.0,0.0),
                   eCycle=0}

initCtlrInput :: CtlrInput
initCtlrInput = CI 0.0 0.0 0

initCtlrState :: CtlrState
initCtlrState = CS False

{-# INLINE divProt #-}
divProt :: Double -> Double
divProt r
     | r < 1e-45 = 1e-45
     | otherwise = r

--           base      move1       move2     result
addVecs :: Point3D -> Point3D -> Point3D -> Point3D
addVecs (xb,yb,zb) (x1,y1,z1) (x2,y2,z2) = 
                      ((xb + (x1-xb) + (x2-xb))
                      ,(yb + (y1-yb) + (y2-yb))
                      ,(zb + (z1-zb) + (z2-zb)))

radToDeg r = r * 360.0 / (2.0 * pi)

degToRad d = d * (2.0 * pi) / 360.0

--------------------------
-- Controller Functions --
--------------------------



-- Tony will fill this in later

controlRobot :: CtlrInput -> CtlrState -> (Double,Double,CtlrState)
             --                             vel   dTheta  newState
controlRobot CI {ciFront=front,ciSide=side,ciClk=clk}
             cs@CS {x1337=x} =

  -- very simple .. const vel = 0.016
  -- if front sensor sees obstruction it
  -- increases theta by two degrees (left turn)

  (0.02, ((2.0 * (frontHandler front)) + (sideHandler side)) , cs) 

  where 

    frontHandler front 
          | front < 0.9 = 10.0
          | otherwise = 0.0

    sideHandler side
          | side  < 0.6 = 5.0
          | side  > 0.7 && side < 0.9 = (-5.0)
          | otherwise = 0.0


--------------------------------
--       Model Updates        --
--------------------------------

-- Notes:
--       Currently internal wall equations are,
--
--                   x = 2.5, x = -2.5,
--                   y = 2.5, y = -2.5
--
--       The robot 'width' and 'height' are 0.4
--       This means that the robot will collide
--       with the wall if it's x or y value is
--       outside of the range,
--
--                   -2.3 <= r <= 2.3
--

updateSensor :: Point3D -> (Double,Double) -> Double
updateSensor (x,y,_) (theta,_) = minimum $ findWallDs
  where
    
    thetaMod = mod' theta 360.0

    findWallDs :: [Double]
    findWallDs = map extractDists 
                       (filter validWalls
                          [(c1,topW),(c2,botW),(c3,leftW),(c4,rightW)])

    extractDists (_,w) = abs w
    validWalls   (c,_) = c thetaMod

    c1 a = a >   0.0 && a < 180.0
    c2 a = a < 360.0 && a > 180.0
    c3 a = a >  90.0 && a < 270.0
    c4 a = a > 270.0 || a <  90.0

    thetaRad = degToRad thetaMod

    rightW = (2.5-x)    / (cos thetaRad)
    topW   = (2.5-y)    / (sin thetaRad)
    leftW  = ((-2.5)-x) / (cos thetaRad)
    botW   = ((-2.5)-y) / (sin thetaRad)

updateFrontSensor :: Point3D -> (Double,Double) -> Double
updateFrontSensor pos dir = updateSensor pos dir

updateSideSensor :: Point3D -> (Double,Double) -> Double
updateSideSensor pos (theta,_) = updateSensor pos (theta+270.0,0.0)

updatePosition :: Point3D -> (Double,Double)  ->  Double  ->  Point3D
              --   oldPos         theta          velocity      newPos
updatePosition (x,y,_) (theta,_) vel =
     -- for now 2D only
     (newX,newY,0.0)
  where
      newX = x + (vel * cos (degToRad theta)) 
      newY = y + (vel * sin (degToRad theta))
  